

#include "internetaddress.h"
#include "tdsocket.h"

#include <netdb.h>
#include <netinet/in.h>


// namespace Coral {


// Print note, if IPv6 is not available.
// #ifdef PRINT_NOIPV6_NOTE


// Check, if IPv6 is available on this host.
bool InternetAddress::UseIPv6 = checkIPv6();


// ###### Internet address constructor ######################################
InternetAddress::InternetAddress()
{
   reset();
   Valid = false;
}


// ###### Internet address constructor ######################################
// ## address: <Name>:port
// ##          <Address>:port
// ##          <Name>            => implies port = 0
// ##          <Address>         => implies port = 0
InternetAddress::InternetAddress(const String& address)
{
   if(address.isNull()) {
      reset();
   }
   else {
      // ====== Handle addresses with additional hostname:port ==============
      // Skip hostname:port, only use (address:port) within brackets
      String host = address;
      String port = "0";
      integer p1 = host.index('(');
      if(p1 > 0) {
         host = host.mid(p1 + 1);
         host = host.left(host.length());
      }

      // ====== Handle RFC 2732 compliant address ===========================
      if(host[0] == '[') {
         p1 = host.index(']');
         if(p1 > 0) {
            if((host[p1 + 1] == ':') || (host[p1 + 1] == '!')) {
               port = host.mid(p1 + 2);
            }
            host = host.mid(1,p1 - 1);
            host = host.left(host.length());
         }
         else {
            Valid = false;
            return;
         }
      }

      // ====== Handle standard address:port ================================
      else {
         p1 = address.rindex(':');
         if(p1 < 0) p1 = address.rindex('!');
         if(p1 > 0) {
            host = address.left(p1);
            port = address.mid(p1 + 1);
         }
      }

      // ====== Initialize address ==========================================
      int portNumber;
      if((sscanf(port.getData(),"%d",&portNumber) == 1) &&
         (portNumber >= 0) &&
         (portNumber <= 65535)) {
         init(host.getData(),portNumber);
      }
      else {
         Valid = false;
      }
   }
}


// ###### Internet address constructor ######################################
InternetAddress::InternetAddress(const InternetAddress& address)
{
   init(address);
}


// ###### Internet address constructor ######################################
InternetAddress::InternetAddress(const String& hostName,
                                 const card16  port)
{
   if(hostName.isNull())
      init(port);
   else
      init(hostName,port);
}


// ###### Internet address constructor ######################################
InternetAddress::InternetAddress(const PortableAddress& address)
{
   for(cardinal i = 0;i < 8;i++) {
      Host[i] = address.Host[i];
   }
   Port  = address.Port;
   Valid = true;
   setPrintFormat(PF_Full);
}


// ###### Internet address constructor ######################################
InternetAddress::~InternetAddress() 
{
}


// ###### Internet address constructor ######################################
InternetAddress::InternetAddress(const card16 port)
{
   init(port);
}


// ###### Internet address constructor ######################################
InternetAddress::InternetAddress(sockaddr* address, const socklen_t length)
{
   init(address,length);
}


// ###### Reset #############################################################
void InternetAddress::reset()
{
   for(cardinal i = 0;i < 8;i++) {
      Host[i] = 0x0000;
   }
   Valid = true;
   setPort(0);
   setPrintFormat(PF_Full);
}


// ###### Check, if address is valid ########################################
bool InternetAddress::isValid() const
{
   return(Valid);
}


// ###### Internet address constructor ######################################
void InternetAddress::init(const InternetAddress& address)
{
   Port = address.Port;
   for(cardinal i = 0;i < 8;i++) {
      Host[i] = address.Host[i];
   }
   Valid = address.Valid;
   setPrintFormat(address.getPrintFormat());
}


// ###### Internet address initializer ######################################
void InternetAddress::init(const card16 port)
{
   for(cardinal i = 0;i < 8;i++) {
      Host[i] = 0x0000;
   }
   Valid = true;
   setPort(port);
   setPrintFormat(PF_Full);
}


// ###### Initialize internet address with hostname and port ################
void InternetAddress::init(const String& hostName, const card16 port)
{
   card16   address[8];
   cardinal length = getHostByName(hostName.getData(),(card16*)&address);

   Valid = true;
   setPort(port);
   setPrintFormat(PF_Full);
   switch(length) {
      case 4:
          for(cardinal i = 0;i < 5;i++) {
             Host[i] = 0x0000;
          }
          Host[5] = 0xffff;
          memcpy((char*)&Host[6],&address,4);
        break;
      case 16:
         memcpy((char*)&Host,(char*)&address,16);
       break;
      default:
        reset();
        Valid = false;
       break;
   }
}


// ###### Internet address initializer ######################################
void InternetAddress::init(const PortableAddress& address)
{
   for(cardinal i = 0;i < 8;i++) {
      Host[i] = address.Host[i];
   }
   Port  = address.Port;
   Valid = true;
   setPrintFormat(PF_Full);
}


// ###### Initialize internet address from system's sockaddr structure ######
void InternetAddress::init(sockaddr* address, const socklen_t length)
{
   setSystemAddress(address,length);
   setPrintFormat(PF_Full);
}


// ###### Get local address #################################################
InternetAddress InternetAddress::getLocalAddress(const InternetAddress& peer)
{
   InternetAddress address;
   Socket socket(Socket::IP,Socket::UDP);
   if(socket.connect(peer)) {
      socket.getSocketAddress(address);   
   }
   return(address);
}


// ###### Get address string ################################################
String InternetAddress::getAddressString() const
{
   // ====== Check for invalid address ======================================
   if(!Valid) {
      return(String("(invalid)"));
   }

   // ====== Initialize =====================================================
   char hostString[256];
   char addressString[256];
   hostString[0]    = 0x00;
   addressString[0] = 0x00;
   if(!((Format & PF_Hostname) || (Format & PF_Address))) {
      cerr << "WARNING: InternetAddress::getAddressString() - "
              "Set PF_Hostname or PF_Address before printing!" << endl;
      return(String("(check print format)"));
   }

   // ====== Generate hostname string =======================================
   if(Format & PF_Hostname) {
      sockaddr_in6 socketAddressBuffer;
      sockaddr*    socketAddress = (sockaddr*)&socketAddressBuffer;
      hostent      resultBuffer;
      hostent*     hostInfo = NULL;
      char         buffer[2048]; 
      int          e;
      if((UseIPv6) || (!isIPv4())) {
         const cardinal addressType    = AF_INET6;
         const cardinal addressLength  = 16;
         const char*    addressPointer =
             (char*)&((sockaddr_in6*)socketAddress)->sin6_addr;
         const cardinal socketAddressLength =
             getSystemAddress(socketAddress,sizeof(sockaddr_in6),AF_INET6);
         if(socketAddressLength > 0) {
            gethostbyaddr_r(addressPointer,addressLength,addressType,
                            &resultBuffer,
                            (char*)&buffer,sizeof(buffer),
                            &hostInfo,&e);
         }
      }
      if(hostInfo == NULL) {
         const cardinal addressType    = AF_INET;
         const cardinal addressLength  = 4;
         const char*    addressPointer =
             (char*)&((sockaddr_in*)socketAddress)->sin_addr;
         const cardinal socketAddressLength =
             getSystemAddress(socketAddress,sizeof(sockaddr_in),AF_INET);
         if(socketAddressLength > 0) {
            gethostbyaddr_r(addressPointer,addressLength,addressType,
                            &resultBuffer,
                            (char*)&buffer,sizeof(buffer),
                            &hostInfo,&e);
         }
      }

      if(hostInfo != NULL) {
         if(!(Format & PF_IPOnly)) {
            snprintf((char*)&hostString,sizeof(hostString),"%s:%d",hostInfo->h_name,ntohs(Port));
         }
         else {
            snprintf((char*)&hostString,sizeof(hostString),"%s",hostInfo->h_name);
         }
      }
   }

   // ====== Generate address string ========================================
   if((Format & PF_Address) || ((Format & PF_Hostname) && (hostString[0] == 0x00))) {
      if((UseIPv6) || (!isIPv4())) {
         char str[32];
         bool compressed = false;

         // If IPv6 address has an embedded IPv4 address
         // print ::ffff:a.b.c.d format
         integer length = isIPv6() ? 8 : 6;

         // Use RFC 2732 compilant notation
         if(!(Format & PF_IPOnly)) {
            strcat((char*)&addressString,"[");
         }
         const cardinal l0 = strlen(addressString);

         // Print IPv6 hex notation
         for(cardinal i = 0;i < (cardinal)length;i++) {
            card16 value = ntohs(Host[i]);
            if((value != 0) || (compressed == true) || (i == 7)) {
               snprintf((char*)&str,sizeof(str),"%x",value);
            }
            else {
               cardinal j;
               for(j = i + 1;j < 8;j++) {
                  if(Host[j] != 0) {
                     break;
                  }
               }
               if(j == i + 1) {
                  snprintf((char*)&str,sizeof(str),"%x",value);
               }
               else {
                  if(i == 0)
                     strcpy((char*)&str,":");
                  else
                     strcpy((char*)&str,"");
                  compressed = true;
                  i = j - 1;
               }                        
            }
            strcat((char*)&addressString,(char*)&str);
            if(i < 7) {
               strcat((char*)&addressString,":");
            }
         }
         if(addressString[l0 + 1] == 0x00) { // Any-Adress "::"
            strcat((char*)&addressString,":");
         }
               
         // Print embedded IPv4 address in IPv4 notation
         if(length == 6) {
            card32 a = ntohl(*((card32*)&Host[6]));
            snprintf((char*)&str,sizeof(str),
                     "%d.%d.%d.%d",(a & 0xff000000) >> 24,
                                   (a & 0x00ff0000) >> 16,
                                   (a & 0x0000ff00) >> 8,
                                   (a & 0x000000ff));
            strcat((char*)&addressString,(char*)&str);
         }
               
         // Add port number
         if(!(Format & PF_IPOnly)) {
            snprintf((char*)&str,sizeof(str),"]:%d",ntohs(Port));
            strcat((char*)&addressString,(char*)&str);
         }
      }
      else {
         card32 a = ntohl(*((card32*)&Host[6]));
         if(!(Format & PF_IPOnly)) {
            snprintf((char*)&addressString,sizeof(addressString),
                     "%d.%d.%d.%d:%d",(a & 0xff000000) >> 24,
                                      (a & 0x00ff0000) >> 16,
                                      (a & 0x0000ff00) >> 8,
                                      (a & 0x000000ff),
                                      ntohs(Port));
         }
         else {
            snprintf((char*)&addressString,sizeof(addressString),
                     "%d.%d.%d.%d",(a & 0xff000000) >> 24,
                                   (a & 0x00ff0000) >> 16,
                                   (a & 0x0000ff00) >> 8,
                                   (a & 0x000000ff));
         }
      }
   }

   if((hostString[0] != 0x00) && (addressString[0] != 0x00))
      return(String(hostString) + " (" + String(addressString) + ")");

   return(String(hostString) + String(addressString));
}


// ###### Get sockaddr structure from internet address ######################
cardinal InternetAddress::getSystemAddress(sockaddr*       buffer,
                                           const socklen_t length,
                                           const cardinal  type) const
{
   switch(type) {
      case AF_INET6: {
         sockaddr_in6* address = (sockaddr_in6*)buffer;
         if(sizeof(sockaddr_in6) <= length) {
            address->sin6_family   = AF_INET6;
            address->sin6_flowinfo = 0;
            address->sin6_port     = Port;
            memcpy((char*)&address->sin6_addr.s6_addr16[0],(char*)&Host,16);
            return(sizeof(sockaddr_in6));
         }
         else {
            cerr << "WARNING: InternetAddress::getSystemInternetAddress() - "
                    "Buffer size too low for AF_INET6!" << endl;
         }
        }
       break;
      case AF_INET: {
         sockaddr_in* address = (sockaddr_in*)buffer;
         if(sizeof(sockaddr_in) <= length) {
             address->sin_family = AF_INET;
            if(isIPv4()) {
               address->sin_port = Port;
               memcpy((char*)&address->sin_addr.s_addr,(char*)&Host[6],4);
               return(sizeof(sockaddr_in));
            }
            else {
               return(0);
            }
         }
         else {
            cerr << "WARNING: InternetAddress::getSystemInternetAddress() - "
                    "Buffer size too low for AF_INET!" << endl;
         }
        }
       break;
      default:
         cerr << "WARNING: InternetAddress::getSystemInternetAddress() - Unknown type "
              << type << "!" << endl;
       break;
   }
   return(0);
}


// ###### Initialize internet address from sockaddr structure ###############
bool InternetAddress::setSystemAddress(sockaddr* address, const socklen_t length)
{
   sockaddr_in* address4 = (sockaddr_in*)address;
   Port = address4->sin_port;

   switch(address4->sin_family) {
      case AF_INET:
         for(cardinal i = 0;i < 5;i++) {
            Host[i] = 0x0000;
         }
         Host[5] = 0xffff;
         memcpy((char*)&Host[6],(char*)&address4->sin_addr.s_addr,4);
         Valid = true;
         return(true);
       break;
      case AF_INET6: {
         sockaddr_in6* address6 = (sockaddr_in6*)address;
         memcpy((char*)&Host,(char*)&address6->sin6_addr.s6_addr16[0],16);
         Valid = true;
         return(true);
        }
       break;
      default:
         reset();
         Valid = true;
        break;
   }
   return(false);
}


// ###### Check, if IPv6 is available #######################################
bool InternetAddress::checkIPv6()
{
   int result = socket(AF_INET6,SOCK_DGRAM,0);
   if(result != -1) {
      close(result);
      _res.options |= RES_USE_INET6;
      return(true);
   }
   else {
#ifdef PRINT_NOIPV6_NOTE
      cerr << "IPv6 is unsupported on this host!" << endl;
#endif
      return(false);
   }      
}


// ###### Get host address by name ##########################################
cardinal InternetAddress::getHostByName(const String& hostName, card16* myadr)
{
   // ====== Check for null address =========================================
   if(hostName.isNull()) {
      for(cardinal i = 0;i < 8;i++) myadr[i] = 0;
      if(UseIPv6)
         return(16);
      else
         return(4);
   }   

   // ====== Get hostent for host ===========================================
   int type = AF_INET;
   if(UseIPv6) type = AF_INET6;   
   hostent  resultBuffer;
   hostent* hostInfo;
   char     buffer[2048]; 
   int      e;
   gethostbyname2_r(hostName.getData(),type,
                    &resultBuffer,
                    (char*)&buffer,sizeof(buffer),
                    &hostInfo,&e);
   if((hostInfo == NULL) && (type == AF_INET6)) {
      gethostbyname2_r(hostName.getData(),AF_INET,
                       &resultBuffer,
                       (char*)&buffer,sizeof(buffer),
                       &hostInfo,&e);
   }
   if((hostInfo == NULL) && (type == AF_INET)) {
      gethostbyname2_r(hostName.getData(),AF_INET6,
                       &resultBuffer,
                       (char*)&buffer,sizeof(buffer),
                       &hostInfo,&e);
   }
   if(hostInfo == NULL)
      return(0);

   // ====== If IPv4 only: Convert IPv6 to IPv4 =============================
   if((!UseIPv6) && (hostInfo->h_addrtype == AF_INET6)) {
      card32* address = (card32*)hostInfo->h_addr_list[0];
      if((address[0] == 0) && (address[1] == 0) && (address[2] == htonl(0x0000ffff))) {
         memcpy((void*)myadr,(void*)&address[3],4);
         return(4);
      }
   }   
   // ====== No conversion necessary ========================================
   memcpy((void*)myadr,(void*)hostInfo->h_addr_list[0],hostInfo->h_length);   
   return(hostInfo->h_length);
}


// }
