

#ifndef SCTPASSOCIATION_H
#define SCTPASSOCIATION_H


#include "tdsystem.h"
#include "condition.h"
#include "internetaddress.h"
#include "sctpsocket.h"
#include "sctpchunkarrivalqueue.h"


#include <sctp.h>


// namespace Coral {


/**
  * This class manages a SCTP assocation. Note: The constructor is protected,
  * a SCTP assocation can be created from an SCTP socket using associate() or
  * accept().
  *
  * @short   SCTP Association
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  *
  * @see SCTPSocket
  * @see SCTPSocket#associate
  * @see SCTPSocket#accept
  */
class SCTPAssociation
{
   // ====== Friend classes =================================================
   friend class SCTPSocket;
   friend class SCTPConnectionlessSocket;
   friend class SCTPSocketMaster;


   // ====== Destructor =====================================================
   public:
   /**
     * Destructor.
     */
   ~SCTPAssociation();


   // ====== Association functions ==========================================
   /**
     * Get internal association ID.
     *
     * @return Association ID.
     */
   unsigned int getID() const;

   /**
     * Get local addresses.
     *
     * @param addressArray Array of addressEntries entries to store local addresses to (NULL to skip).
     * @param addressEntries Number of addressEntries; this will be overwritten by the number of local addresses.
     * @return true, if addressEntries are sufficient; false otherwise.
     */
   bool getLocalAddresses(InternetAddress** addressArray,
                          unsigned int&     addressEntries);

   /**
     * Get remote addresses.
     *
     * @param addressArray Array of addressEntries entries to store remote addresses to (NULL to skip).
     * @param addressEntries Number of addressEntries; this will be overwritten by the number of remote addresses.
     * @return true, if addressEntries are sufficient; false otherwise.
     */
   bool getRemoteAddresses(InternetAddress** addressArray,
                           unsigned int&     addressEntries);


   /**
     * @param dataBuffer Buffer to store data to.
     * @param dataBufferSize Size of data buffer; this will be overwritten with actual size of data content.
     * @param controlBuffer Buffer to store control data to.
     * @param controlBufferSize Size of control buffer; this will be overwritten with actual size of control content.
     * @param flags Flags; this will be overwritten with actual reception flags.
     * @param streamID Variable to store stream ID to.
     * @param protoID Variable to store protocol ID to.
     * @return error code (0 for success).
     */
   int receive(char*           dataBuffer,
               size_t&         dataBufferSize,
               char*           controlBuffer,
               size_t&         controlBufferSize,
               int&            flags,
               unsigned short& streamID,
               unsigned int&   protoID);

   /**
     * Send data.
     *
     * @param buffer Data to be sent.
     * @param length Length of data to be sent.
     * @param flags Flags.
     * @param streamID Stream ID.
     * @param protoID Protocol ID.
     * @return error code (0 for success).
     */
   int send(const char*          buffer,
            const size_t         length,
            const int            flags,
            const unsigned short streamID,
            const unsigned int   protoID);


   /**
     * Shutdown.
     */
   void shutdown();

   /**
     * Abort.
     */
   void abort();


   // ====== Protected data =================================================
   protected:
   SCTPAssociation(SCTPSocket* socket, const unsigned int associationID);


   // ====== Private data ===================================================
   private:
   SCTPSocket*       Socket;
   ChunkArrivalQueue InQueue;
   Condition         EstablishCondition;
   Condition         ShutdownCompleteCondition;

   unsigned int      AssociationID;

   bool              CommunicationLostNotification;
   bool              ShutdownCompleteNotification;
   bool              IsShuttingDown;
};


// }


#endif
