

#ifndef CHUNKARRIVALQUEUE_H
#define CHUNKARRIVALQUEUE_H


#include "tdsystem.h"
#include "condition.h"

#include <sctp.h>


// namespace Coral {



/**
  * Association status.
  *
  * @short   Association Status
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
struct AssociationStatus
{
   /**
     * Association ID.
     */
   unsigned int AssocID;

   /**
     * Stream ID.
     */
   unsigned short StreamID;

   /**
     * Remote port.
     */
   unsigned short RemotePort;

   /**
     * Number of remote addresses.
     */
   unsigned short RemoteAddresses;

   /**
     * Remote addresses.
     */
   char RemoteAddress[SCTP_MAX_NUM_ADDRESSES][SCTP_MAX_IP_LEN];

   /**
     * SCTP Association Status.
     */
   SCTP_Association_Status Status;

   /**
     * SCTP Path Status for each remote address.
     */
   SCTP_Path_Status PathStatus[SCTP_MAX_NUM_ADDRESSES];
};


/**
  * Association control data.
  *
  * @short   Association Control Data
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
struct AssociationControlData
{
   /**
     * Control data types.
     */
   enum AssociationControlDataType {
      ACDT_Unknown             = 0,
      ACDT_CommunicationUp     = 1,
      ACDT_Restart             = 2,
      ACDT_NetworkStatusChange = 3,
      ACDT_CommunicationError  = 4,
      ACDT_SendFailure         = 5,
      ACDT_CommunicationLost   = 6,
      ACDT_ShutdownComplete    = 7
   };

   /**
     * Type of control data.
     */
   unsigned int Type;
};



/**
  * This class is a queue for SCTP Data Arrive notification data.
  *
  * @short   Chunk Arrival Queue
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
class ChunkArrivalQueue
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   ChunkArrivalQueue();

   /**
     * Destructor.
     */
   ~ChunkArrivalQueue();


   // ====== Queuing functions ==============================================
   /**
     * Add chunk to tail of queue. Note: The controlData pointer itself is
     * stored. No copy will be made!
     *
     * @param assocID Association ID.
     * @param streamID Stream ID.
     * @param protoID Protocol ID.
     * @param length Length of chunk.
     * @param unordered Unordered flag.
     * @param controlData Pointer to control data.
     * @param controlDataLength Length of control data.
     * @param status Association status.
     * @return true, if add was successful; false otherwise (out of memory).
     */
   bool addChunk(const unsigned int       assocID,
                 const unsigned short     streamID,
                 const unsigned int       protoID,
                 const unsigned int       length,
                 const unsigned int       unordered,
                 AssociationControlData*  controlData,
                 const unsigned int       controlDataLength,
                 const AssociationStatus& status);

   /**
     * Add chunk to top of queue. Note: The controlData pointer itself is
     * stored. No copy will be made!
     *
     * @param assocID Association ID.
     * @param streamID Stream ID.
     * @param protoID Protocol ID.
     * @param length Length of chunk.
     * @param unordered Unordered flag.
     * @param controlData Pointer to control data.
     * @param controlDataLength Length of control data.
     * @param status Association status.
     * @return true, if push was successful; false otherwise (out of memory).
     */
   bool pushChunk(const unsigned int       assocID,
                  const unsigned short     streamID,
                  const unsigned int       protoID,
                  const unsigned int       length,
                  const unsigned int       unordered,
                  AssociationControlData*  controlData,
                  const unsigned int       controlDataLength,
                  const AssociationStatus& status);

   /**
     * Get chunk from top of queue and remove it. Note: The control data
     * has to be freed by the user of this function!
     *
     * @param assocID Reference to store Association ID.
     * @param streamID Reference to store Stream ID.
     * @param protoID Reference to store Protocol ID.
     * @param length Reference to store Length of chunk.
     * @param unordered Reference to store Unordered flag.
     * @param controlData Reference to store Pointer to control data.
     * @param controlDataLength Reference to store Length of control data.
     * @param status Reference to store Association status.
     * @return true, if get was successful; false otherwise (queue is empty).
     */
   bool getChunk(unsigned int&            assocID,
                 unsigned short&          streamID,
                 unsigned int&            protoID,
                 unsigned int&            length,
                 unsigned int&            unordered,
                 AssociationControlData*& controlData,
                 unsigned int&            controlDataLength,
                 AssociationStatus&       status);

   /**
     * Wait for new chunk.
     */
   void waitForChunk();

   /**
     * Signalize, that new chunk has arrived.
     */
   void signal();

   /**
     * Flush queue.
     */
   void flush();


   // ====== Private data ===================================================
   private:
   struct Chunk
   {
      Chunk*                  NextChunk;
      unsigned int            AssociationID;
      unsigned short          StreamID;
      unsigned int            ProtoID;
      unsigned int            Length;
      unsigned int            Unordered;
      unsigned int            ControlDataLength;
      AssociationControlData* ControlData;
      AssociationStatus       Status;
   };

   Chunk*    FirstChunk;
   Chunk*    LastChunk;
   Condition DataArriveCondition;
};


// }


#endif
