

#ifndef SCTPCONNECTIONLESSSOCKET_H
#define SCTPCONNECTIONLESSSOCKET_H


#include "tdsystem.h"
#include "sctpsocket.h"
#include "sctpassociation.h"


#include <sctp.h>
#include <multimap.h>


// namespace Coral {


/**
  * This class manages a connectionless (UDP-style) SCTP socket. Internally,
  * this class automatically establishs or accepts new SCTP associations.
  *
  * @short   SCTP Connectionless Socket
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
class SCTPConnectionlessSocket : public SCTPSocket
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     *
     * @param server true to accept new incoming associations (receiver mode); false otherwise (sender mode, default).
     */
   SCTPConnectionlessSocket(const bool server = false);

   /**
     * Destructor.
     */
   ~SCTPConnectionlessSocket();


   // ====== SCTP Socket functions ==========================================
   /**
     * Receive data.
     *
     * @param dataBuffer Buffer to store data to.
     * @param dataBufferSize Size of data buffer; this will be overwritten with actual size of data content.
     * @param controlBuffer Buffer to store control data to.
     * @param controlBufferSize Size of control buffer; this will be overwritten with actual size of control content.
     * @param flags Flags; this will be overwritten with actual reception flags.
     * @param streamID Variable to store stream ID to.
     * @param protoID Variable to store protocol ID to.
     * @param addressArray Array of addressEntries entries to store the sender's addresses to (NULL to skip).
     * @param addressEntries Number of addressEntries; this will be overwritten by the sender's number of peer addresses.
     * @param status Variable to store AssocationStatus data to.
     * @return error code (0 for success).
     */
   int receiveFrom(char*              dataBuffer,
                   size_t&            dataBufferSize,
                   char*              controlBuffer,
                   size_t&            controlBufferSize,
                   int&               flags,
                   unsigned short&    streamID,
                   unsigned int&      protoID,
                   InternetAddress**  addressArray,
                   unsigned int&      addressEntries,
                   AssociationStatus& status);

   /**
     * Send data.
     *
     * @param buffer Data to be sent.
     * @param length Length of data to be sent.
     * @param flags Flags.
     * @param streamID Stream ID.
     * @param protoID Protocol ID.
     * @param destinationAddress Destination address.
     * @return error code (0 for success).
     */
   int sendTo(const char*            buffer,
              const size_t           length,
              const int              flags,
              const unsigned short   streamID,
              const unsigned int     protoID,
              const InternetAddress& destinationAddress);


   // ====== Private data ===================================================
   private:
   multimap<unsigned int, SCTPAssociation*> OutgoingAssociationList;
};


// }


#endif
