

#include "sctpsocket.h"
#include "sctpsocketmaster.h"


// namespace Coral {


#define DEBUG
#define PRINT_BIND
#define PRINT_ACCEPT
#define PRINT_ASSOCIATE


// ###### Constructor #######################################################
SCTPSocket::SCTPSocket(const bool hasGlobalQueue)
{
   InstanceName       = 0;
   ConnectionRequests = NULL;
   HasGlobalQueue     = hasGlobalQueue;
}


// ###### Destructor ########################################################
SCTPSocket::~SCTPSocket()
{
   unbind();
}


// ###### Get association for association ID ################################
SCTPAssociation* SCTPSocket::getAssociationForAssociationID(const unsigned int assocID,
                                                            const bool activeOnly)
{
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      SCTPAssociationList.find(assocID);
   if(iterator != SCTPAssociationList.end()) {
      if(!((iterator->second->IsShuttingDown) && (activeOnly))) {
         return(iterator->second);
      }
   }
   return(NULL);
}


// ###### Get local addresses ###############################################
bool SCTPSocket::getLocalAddresses(InternetAddress** addressArray,
                                   unsigned int&     addressEntries)
{
   unsigned int i;
   bool         result = true;

   SCTPSocketMaster::MasterInstance.lock();
   for(i = 0;i < NoOfLocalAddresses;i++) {
      if(i >= addressEntries) {
         result = false;
         break;
      }
      addressArray[i]->init((char*)LocalAddressList[i],LocalPort);
   }
   addressEntries = i;
   SCTPSocketMaster::MasterInstance.unlock();

   return(result);
}


// ###### Bind socket to local address(es) and port #########################
int SCTPSocket::bind(const unsigned short    localPort,
                     const unsigned short    noOfInStreams,
                     const unsigned short    noOfOutStreams,
                     const unsigned int      noOfLocalAddresses,
                     const InternetAddress** localAddressList)
{
   if(SCTPSocketMaster::InitializationResult != 0) {
      cerr << "WARNING: SCTPSocket::bind() - SCTP is not initialized!" << endl;
      return(-EPROTONOSUPPORT);
   }
   SCTPSocketMaster::MasterInstance.lock();

   // ====== Initialize =====================================================
   unbind();
   LocalPort          = localPort;
   NoOfInStreams      = noOfInStreams;
   NoOfOutStreams     = noOfOutStreams;
   NoOfLocalAddresses = noOfLocalAddresses;
   NoOfLocalAddresses = 0;


   // ====== Initialize local addresses =====================================
#ifdef PRINT_BIND
   cout << "Binding to {";
   for(unsigned int i = 0;i < noOfLocalAddresses;i++) {
      InternetAddress address(*(localAddressList[i]));
      address.setPrintFormat(InternetAddress::PF_Address);
      address.setPort(localPort);
      cout << " " << address << " ";
   }
   cout << "}." << endl;
#endif
   for(unsigned int i = 0;i < min(noOfLocalAddresses,(unsigned int)SCTP_MAX_NUM_ADDRESSES);i++) {
      InternetAddress address(*(localAddressList[i]));
      address.setPrintFormat(InternetAddress::PF_IPOnly|InternetAddress::PF_Address);
      snprintf((char*)&LocalAddressList[i],SCTP_MAX_IP_LEN,"%s",address.getAddressString().getData());
      NoOfLocalAddresses++;
   }
   if(NoOfLocalAddresses < 1) {
      cerr << "ERROR: SCTPSocket::bind() - No local addresses!" << endl;
      SCTPSocketMaster::MasterInstance.unlock();
      return(-EINVAL);
   }


   // ====== Register SCTP instance =========================================
   InstanceName = sctp_registerInstance(LocalPort,NoOfInStreams,NoOfOutStreams,
                                        NoOfLocalAddresses,LocalAddressList,
                                        SCTPSocketMaster::SCTPCallbacks);
   if(InstanceName == 0) {
      cerr << "ERROR: SCTPSocket::bind() - sctp_registerInstance() failed!" << endl;
      SCTPSocketMaster::MasterInstance.unlock();
      return(-EFAULT);
   }


   // ====== Add socket to global list ======================================
   SCTPSocketMaster::SCTPSocketList.insert(pair<unsigned short, SCTPSocket*>(InstanceName,this));


   SCTPSocketMaster::MasterInstance.unlock();
   return(0);
}


// ###### Release socket binding ############################################
void SCTPSocket::unbind()
{
   if(InstanceName != 0) {
      SCTPSocketMaster::MasterInstance.lock();

      // ====== Remove socket from global list ==============================
      multimap<unsigned short, SCTPSocket*>::iterator socketIterator =
         SCTPSocketMaster::SCTPSocketList.find(InstanceName);
      if(socketIterator != SCTPSocketMaster::SCTPSocketList.end()) {
         SCTPSocketMaster::SCTPSocketList.erase(socketIterator);
      }

      // ====== Remove all associations =====================================
      multimap<unsigned int, SCTPAssociation*>::iterator associationIterator =
         SCTPAssociationList.begin();
      while(associationIterator != SCTPAssociationList.end()) {
         delete associationIterator->second;
         SCTPAssociationList.erase(associationIterator);
         associationIterator = SCTPAssociationList.begin();
      }

      // ====== Unregister SCTP instance ====================================
      sctp_unregisterInstance(InstanceName);
      InstanceName = 0;

      SCTPSocketMaster::MasterInstance.unlock();

      GlobalQueue.flush();
      ServerMode = false;
   }
}


// ###### Set listen mode ###################################################
void SCTPSocket::listen(const unsigned int backlog)
{
   SCTPSocketMaster::MasterInstance.lock();
   ServerMode = true;
   SCTPSocketMaster::MasterInstance.unlock();
}


// ###### Accept new association ############################################
SCTPAssociation* SCTPSocket::accept(InternetAddress** addressArray,
                                    unsigned int&     addressEntries,
                                    const bool        wait)
{
   SCTPSocketMaster::MasterInstance.lock();
   if(!ServerMode) {
      cerr << "ERROR: SCTPSocket::accept() - Socket is not in server mode, call listen() first!" << endl;
      exit(1);
   }


   // ====== Get new association=============================================
   while(ConnectionRequests == NULL) {
      if(!wait) {
         return(NULL);
      }
      SCTPSocketMaster::MasterInstance.unlock();
      EstablishCondition.wait();
      SCTPSocketMaster::MasterInstance.lock();
   }
   if(ConnectionRequests == NULL) {
      return(NULL);
   }


   // ====== Initialize address array =======================================
   if(addressArray != NULL) {
      unsigned int i;
      for(i = 0;i < ConnectionRequests->Status.RemoteAddresses;i++) {
         if(i >= addressEntries) {
            break;
         }
         addressArray[i]->init(String((char*)&ConnectionRequests->Status.RemoteAddress[i]),ConnectionRequests->Status.RemotePort);
      }
      addressEntries = i;
   }
#ifdef PRINT_ACCEPT
   cout << "Accepted association #" << ConnectionRequests->Association->AssociationID << " from {";
   for(unsigned int i = 0;i < ConnectionRequests->Status.RemoteAddresses;i++) {
      if(i >= addressEntries) {
         break;
      }
      InternetAddress address(String((char*)&ConnectionRequests->Status.RemoteAddress[i]),ConnectionRequests->Status.RemotePort);
      address.setPrintFormat(InternetAddress::PF_Address);
      cout << " " << address << " ";
   }
   cout << "}." << endl;
#endif


   // ====== Remove association from incoming list ==========================
   SCTPAssociation* association   = ConnectionRequests->Association;
   IncomingConnection* oldRequest = ConnectionRequests;
   ConnectionRequests = oldRequest->NextConnection;
   delete oldRequest;


   SCTPSocketMaster::MasterInstance.unlock();
   return(association);
}


// ###### Establish new association #########################################
SCTPAssociation* SCTPSocket::associate(const unsigned short   noOfOutStreams,
                                       const InternetAddress& destinationAddress,
                                       const bool             wait)
{
   InternetAddress destAddress(destinationAddress);
   destAddress.setPrintFormat(InternetAddress::PF_IPOnly|InternetAddress::PF_Address);

   // ====== Establish new association ======================================
   SCTPSocketMaster::MasterInstance.lock();
   const unsigned int assocID = sctp_associate(InstanceName,
                                               noOfOutStreams,
                                               (unsigned char*)destAddress.getAddressString().getData(),
                                               destAddress.getPort(),
                                               NULL);
#ifdef PRINT_ASSOCIATE
   InternetAddress address(destinationAddress);
   address.setPrintFormat(InternetAddress::PF_Address);
   cout << "Association to " << address << " => ID #"
        << assocID << "." << endl;
#endif


   // ====== Create SCTPAssociation object ==================================
   SCTPAssociation* association = NULL;
   if(assocID != 0) {
      association = new SCTPAssociation(this,assocID);
      if(association == NULL) {
         sctp_abort(assocID);
         sctp_deleteAssociation(assocID);
         cerr << "ERROR: SCTPSocket::associate() - Out of memory!" << endl;
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();


   // ====== Wait for peer's connection up notification =====================
   if((association != NULL) && (wait)) {
      if(association->EstablishCondition.timedWait(5000000) == true) {
#ifdef PRINT_ASSOCIATE
         cout << "Association #" << assocID << " established." << endl;
#endif
         return(association);
      }
#ifdef PRINT_ASSOCIATE
         cout << "Association #" << assocID << " failed!" << endl;
#endif
      delete association;
   }

   return(NULL);
}


// ###### Get error code for given association ID ###########################
int SCTPSocket::getErrorCode(const unsigned int assocID)
{
  SCTPAssociation* association = getAssociationForAssociationID(assocID,false);
   if(association != NULL) {
      if(association->ShutdownCompleteNotification) {
         return(-ESHUTDOWN);
      }
      else if(association->CommunicationLostNotification) {
         return(-ECONNABORTED);
      }
   }
   return(0);
}


// ###### Internal receive implementation ###################################
int SCTPSocket::internalReceive(ChunkArrivalQueue& queue,
                                char*              dataBuffer,
                                size_t&            dataBufferSize,
                                char*              controlBuffer,
                                size_t&            controlBufferSize,
                                int&               flags,
                                unsigned int&      assocID,
                                unsigned short&    streamID,
                                unsigned int&      protoID,
                                AssociationStatus& status)
{
   int                     result = 0;
   unsigned int            unordered;
   unsigned int            dataLength;
   unsigned int            controlDataLength;
   AssociationControlData* controlData;


   // ====== Check parameters ===============================================
   if(flags & MSG_WAITALL) {
      cerr << "NOT IMPLEMENTED: SCTPSocket::internalReceive() - MSG_WAITALL flag set!" << endl;
      exit(1);
   }
   if(dataBufferSize == 0) {
      cerr << "WARNING: SCTPSocket::internalReceive() - Data buffer size is zero!" << endl;
      return(0);
   }


   // ====== Get next data or control chunk from queue =====================
   SCTPSocketMaster::MasterInstance.lock();
   bool received = queue.getChunk(assocID,streamID,protoID,dataLength,unordered,controlData,controlDataLength,status);
   if(received == false) {
      // ====== No chunk available -> wait for chunk ======================
      SCTPSocketMaster::MasterInstance.unlock();
      const int errorCode = getErrorCode(assocID);
      if(errorCode != 0) {
         dataBufferSize    = 0;
         controlBufferSize = 0;
         return(errorCode);
      }
      if(flags & MSG_DONTWAIT) {
         dataBufferSize    = 0;
         controlBufferSize = 0;
         return(0);
      }
      queue.waitForChunk();
      SCTPSocketMaster::MasterInstance.lock();
      received = queue.getChunk(assocID,streamID,protoID,dataLength,unordered,controlData,controlDataLength,status);
   }
   SCTPSocketMaster::MasterInstance.unlock();


   // ====== Check, if error occured ========================================
   if(!received) {
      cerr << "INTERNAL ERROR: SCTPSocket::internalReceive() - Nothing received!" << endl;
      exit(1);
   }


   SCTPSocketMaster::MasterInstance.lock();


   // ====== Read data ======================================================
   if(dataLength > 0) {
      unsigned int receivedBytes = min(dataLength,dataBufferSize);
      const int ok = sctp_receive(assocID,streamID,(unsigned char*)dataBuffer,&receivedBytes,(flags & MSG_PEEK) ? SCTP_MSG_PEEK : SCTP_MSG_DEFAULT);
      if(ok == 0) {
         dataBufferSize = receivedBytes;

         // ====== Peek mode: Restore chunk arrival information ===================
         if(flags & MSG_PEEK) {
            queue.pushChunk(assocID,streamID,protoID,dataLength,unordered,controlData,controlDataLength,status);
         }
         else {
            dataLength -= receivedBytes;
            if(dataLength > 0) {
               queue.pushChunk(assocID,streamID,protoID,dataLength,unordered,controlData,controlDataLength,status);
            }
         }
      }
      else {
         cerr << "WARNING: SCTPSocket::internalReceive() - sctp_receive() failed!" << endl;
         result         = -ECONNABORTED;
         dataBufferSize = 0;
      }
   }
   else {
      dataBufferSize = 0;
   }

   // ====== Copy control chunk =========================================
   if((controlBuffer != NULL) && (controlDataLength > 0)) {
      if(controlDataLength > controlBufferSize) {
         cerr << "ERROR: SCTPSocket::internalReceive() - Control buffer too small!" << endl;
         result            = -EFBIG;
         controlBufferSize = 0;
      }
      else {
         controlBufferSize = controlDataLength;
         memcpy(controlBuffer,controlData,controlBufferSize);
      }
   }
   else {
      controlBufferSize = 0;
   }

   SCTPSocketMaster::MasterInstance.unlock();


   // ====== Free control data, if existing =================================
   if(controlData) {
      delete controlData;
   }

   // ====== Update flags ===================================================
   if(unordered == SCTP_UNORDERED_DELIVERY) {
      flags |= MSG_UNORDERED;
   }
   if(dataLength == 0) {
      flags |= MSG_EOR;
   }
   else {
      flags &= ~MSG_EOR;
   }

   return(result);
}


// ###### Internal send implementation ######################################
int SCTPSocket::internalSend(const char*          buffer,
                             const size_t         length,
                             const int            flags,
                             const unsigned int   assocID,
                             const unsigned short streamID,
                             const unsigned int   protoID)
{
   // ====== Check error code ===============================================
   const int errorCode = getErrorCode(assocID);
   if(errorCode != 0) {
      return(errorCode);
   }


   // ====== Do send ========================================================
   SCTPSocketMaster::MasterInstance.lock();
   const int result =
      sctp_send(
         assocID, streamID,
         (unsigned char*)buffer, length,
         protoID,
         SCTP_USE_PRIMARY,
         SCTP_NO_CONTEXT,
         SCTP_INFINITE_LIFETIME,
         ((flags & MSG_UNORDERED) ? SCTP_UNORDERED_DELIVERY : SCTP_ORDERED_DELIVERY),
         ((flags & MSG_UNBUNDLED) ? SCTP_BUNDLING_DISABLED : SCTP_BUNDLING_ENABLED));
   SCTPSocketMaster::MasterInstance.unlock();
   return((result == 0) ? (int)length : -1);
}


// }
