

#ifndef SCTPSOCKET_H
#define SCTPSOCKET_H


#include "tdsystem.h"
#include "condition.h"
#include "internetaddress.h"
#include "sctpassociation.h"
#include "sctpchunkarrivalqueue.h"


#include <sctp.h>
#include <multimap.h>


// namespace Coral {


#define MSG_SHUTDOWN  (1 << 30)
#define MSG_UNORDERED (1 << 29)
#define MSG_UNBUNDLED (1 << 28)


/**
  * This class manages a SCTP socket (SCTP instance).
  *
  * @short   SCTP Socket
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
class SCTPSocket
{
   // ====== Friend classes =================================================
   friend class SCTPSocketMaster;
   friend class SCTPAssociation;


   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     *
     * @param hasGlobalQueue true to use a global queue for all associations; false otherwise (default).
     */
   SCTPSocket(const bool hasGlobalQueue = false);

   /**
     * Destructor.
     */
   ~SCTPSocket();


   // ====== SCTP Socket functions ==========================================
   /**
     * Bind socket to local port and address(es).
     *
     * @param localPort Local port.
     * @param noOfInStreams Number of incoming streams.
     * @param noOfOutStreams Number of outgoing streams.
     * @param noOfLocalAddresses Number of local addresses.
     * @param localAddressList Array of local addresses.
     */
   int bind(const unsigned short    localPort,
            const unsigned short    noOfInStreams,
            const unsigned short    noOfOutStreams,
            const unsigned int      noOfLocalAddresses,
            const InternetAddress** localAddressList);

   /**
     * Release socket binding.
     *
     * @see bind
     */
   void unbind();

   /**
     * Establish new association.
     *
     * @param noOfOutStreams Number of outgoing streams.
     * @param destinationAddress Destination address.
     * @param wait true to wait for establishment (default); false otherwise.
     * @return Association or NULL in case of failure.
     */
   SCTPAssociation* associate(const unsigned short   noOfOutStreams,
                              const InternetAddress& destinationAddress,
                              const bool             wait = true);

   /**
     * Set socket to listen mode: accept new incoming assocations.
     *
     * @param backlog Maximum number of incoming connections to accept simultaneously.
     */
   void listen(const unsigned int backlog);

   /**
     * Wait for incoming association.
     *
     * @param addressArray Array of addressEntries entries to store new association's peer addresses to (NULL to skip).
     * @param addressEntries Number of addressEntries; this will be overwritten by the new association's number of peer addresses.
     * @param wait true to wait for new association (default); false otherwise.
     * @return New association or NULL in case of failure.
     */
   SCTPAssociation* accept(InternetAddress** addressArray,
                           unsigned int&     addressEntries,
                           const bool        wait = true);

   /**
     * Get socket's local addresses.
     *
     * @param addressArray Array of addressEntries entries to store local addresses to (NULL to skip).
     * @param addressEntries Number of addressEntries; this will be overwritten by the number of local addresses.
     * @return true, if addressEntries are sufficient; false otherwise.
     */
   bool getLocalAddresses(InternetAddress** addressArray,
                          unsigned int&     addressEntries);


   // ====== Protected data =================================================
   protected:
   SCTPAssociation* getAssociationForAssociationID(const unsigned int assocID,
                                                   const bool activeOnly = true);
   int getErrorCode(const unsigned int assocID);
   int internalReceive(ChunkArrivalQueue& queue,
                       char*              dataBuffer,
                       size_t&            dataBufferSize,
                       char*              controlBuffer,
                       size_t&            controlBufferSize,
                       int&               flags,
                       unsigned int&      assocID,
                       unsigned short&    streamID,
                       unsigned int&      protoID,
                       AssociationStatus& status);
   int internalSend(const char*          buffer,
                    const size_t         length,
                    const int            flags,
                    const unsigned int   assocID,
                    const unsigned short streamID,
                    const unsigned int   protoID);


   struct IncomingConnection
   {
      IncomingConnection* NextConnection;
      SCTPAssociation*    Association;
      AssociationStatus   Status;
   };

   IncomingConnection*                      ConnectionRequests;
   ChunkArrivalQueue                        GlobalQueue;
   multimap<unsigned int, SCTPAssociation*> SCTPAssociationList;
   bool                                     HasGlobalQueue;
   bool                                     ServerMode;

   unsigned short                           InstanceName;
   unsigned short                           LocalPort;
   unsigned short                           NoOfInStreams;
   unsigned short                           NoOfOutStreams;
   unsigned int                             NoOfLocalAddresses;
   unsigned char                            LocalAddressList[SCTP_MAX_NUM_ADDRESSES][SCTP_MAX_IP_LEN];
   Condition                                EstablishCondition;
};


// }


#endif
