


#include "sctpsocketmaster.h"


// namespace Coral {



#define PRINT_NOTIFICATIONS
// #define PRINT_ARRIVENOTIFICATION



SCTPSocketMaster                      SCTPSocketMaster::MasterInstance;
SCTP_ulpCallbacks                     SCTPSocketMaster::SCTPCallbacks;
multimap<unsigned short, SCTPSocket*> SCTPSocketMaster::SCTPSocketList;
int                                   SCTPSocketMaster::InitializationResult = -1000;


// ###### Constructor #######################################################
SCTPSocketMaster::SCTPSocketMaster()
{
   if(InitializationResult == -1000) {
      SCTPCallbacks.dataArriveNotif          = &dataArriveNotif;
      SCTPCallbacks.sendFailureNotif         = &sendFailureNotif;
      SCTPCallbacks.networkStatusChangeNotif = &networkStatusChangeNotif;
      SCTPCallbacks.communicationUpNotif     = &communicationUpNotif;
      SCTPCallbacks.communicationLostNotif   = &communicationLostNotif;
      SCTPCallbacks.communicationErrorNotif  = &communicationErrorNotif;
      SCTPCallbacks.restartNotif             = &restartNotif;
      SCTPCallbacks.shutdownCompleteNotif    = &shutdownCompleteNotif;
      InitializationResult = sctp_initLibrary();
      if(InitializationResult == 0) {
         start();
      }
      else {
         cerr << "************************************" << endl
              << "WARNING: SCTP initialization failed!" << endl
              << "************************************" << endl;
      }
   }
   else {
      cerr << "ERROR: SCTPSocketMaster::SCTPSocketMaster() - "
              "Do not try to initialice SCTPSocketMaster singleton twice!" << endl;
      exit(1);
   }
}


// ###### Destructor ########################################################
SCTPSocketMaster::~SCTPSocketMaster()
{
}


// ###### SCTP event loop thread ############################################
void SCTPSocketMaster::run()
{
   for(;;) {
      sctp_extendedEventLoop(lock,unlock,(void*)this);
   }
}


// ###### SCTP data arrive notification callback ############################
void SCTPSocketMaster::dataArriveNotif(unsigned int assocID,
                                       unsigned int streamID,
                                       unsigned int length,
                                       unsigned int protoID,
                                       unsigned int unordered,
                                       void*        ulpDataPtr)
{
#ifdef PRINT_ARRIVENOTIFICATION
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d S%02d: Data Arrive Notification, Length = %d",
               assocID,streamID,length);
   cerr << str << endl;
#endif


   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      AssociationStatus status;
      if(SCTPSocketMaster::getAssociationStatus(assocID,streamID,status)) {
         if(socket->HasGlobalQueue) {
            socket->GlobalQueue.addChunk(assocID,streamID,protoID,length,unordered,NULL,0,status);
         }
         else {
            SCTPAssociation* association = socket->getAssociationForAssociationID(assocID);
            if(association != NULL) {
               association->InQueue.addChunk(assocID,streamID,protoID,length,unordered,NULL,0,status);
            }
         }
      }
      else {
         cerr << "INTERNAL ERROR: SCTPSocketMaster::dataArriveNotif() - sctp_getAssocStatus() failure!"
              << endl;
         exit(1);
      }
   }
}


// ###### SCTP send failure notification callback ###########################
void SCTPSocketMaster::sendFailureNotif(unsigned int   assocID,
                                        unsigned char* unsent_data,
                                        unsigned int   dataLength,
                                        unsigned int*  context,
                                        void*          dummy)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Send Failure Notification",assocID);
   cerr << str << endl;
#endif


   // ====== Generate "Send Failure" control chunk =========================
   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_SendFailure;
            AssociationStatus status;
            SCTPSocketMaster::getAssociationStatus(assocID,0,status);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::sendFailureNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }
}


// ###### SCTP network status change notification callback ##################
void SCTPSocketMaster::networkStatusChangeNotif(unsigned int   assocID,
                                                short          destAddrIndex,
                                                unsigned short newState,
                                                void*          ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Network Status Change",assocID);
   cerr << str << endl;
#endif


   // ====== Select new primary path, if it has become inactive ============
   if((newState == SCTP_PATH_UNREACHABLE) && (destAddrIndex == sctp_getPrimary(assocID))) {
      SCTP_AssociationStatus assocStatus;
      SCTP_PathStatus        pathStatus;
      unsigned short         pathID;
      sctp_getAssocStatus(assocID,&assocStatus);
      for(pathID = 0;pathID < assocStatus.numberOfAddresses;pathID++) {
         sctp_getPathStatus(assocID,pathID,&pathStatus);
         if(pathStatus.state == SCTP_PATH_OK) {
            break;
         }
      }
      if(pathID < assocStatus.numberOfAddresses) {
         sctp_setPrimary(assocID, pathID);
      }
   }


   // ====== Generate "Network Status Change" control chunk ================
   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_NetworkStatusChange;
            AssociationStatus status;
            SCTPSocketMaster::getAssociationStatus(assocID,0,status);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::communicationLostNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }
}


// ###### SCTP communication up notification callback #######################
void* SCTPSocketMaster::communicationUpNotif(unsigned int   assocID,
                                             unsigned short status,
                                             int            noOfDestinations,
                                             unsigned short noOfInStreams,
                                             unsigned short noOfOutStreams,
                                             void*          dummy)
{	
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Communication Up Notification",assocID);
   cerr << str << endl;
#endif


   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {

      // ====== Successfull associate() execution ===========================
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID);
      if(association != NULL) {
         association->EstablishCondition.broadcast();
      }

      // ====== Incoming connection =========================================
      else {
         association = new SCTPAssociation(socket,assocID);
         if(association != NULL) {
            SCTPSocket::IncomingConnection* newConnection = new SCTPSocket::IncomingConnection;
            if(newConnection != NULL) {
               newConnection->NextConnection = NULL;
               newConnection->Association    = association;
               SCTPSocketMaster::getAssociationStatus(assocID,0,newConnection->Status);

               if(socket->ConnectionRequests == NULL) {
                  socket->ConnectionRequests = newConnection;
               }
               else {
                  SCTPSocket::IncomingConnection* c = socket->ConnectionRequests;
                  while(c->NextConnection != NULL) {
                     c = c->NextConnection;
                  }
                  c->NextConnection = newConnection;
               }

               socket->EstablishCondition.broadcast();
            }
         }
      }


      // ====== Generate "Communication Up" control chunk ==================
      if(association != NULL) {
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_CommunicationUp;
            AssociationStatus status;
            SCTPSocketMaster::getAssociationStatus(assocID,0,status);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::shutdownCompleteNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }

   return(NULL);
}


// ###### SCTP communication lost notification callback #####################
void SCTPSocketMaster::communicationLostNotif(unsigned int   assocID,
                                              unsigned short status,
                                              void*          ulpDataPtr)
{	
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Communication Lost Notification, Status=%d",assocID,status);
   cerr << str << endl;
#endif


   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         // ====== Notify waiting main thread ===============================
         association->CommunicationLostNotification = true;
         association->ShutdownCompleteNotification  = true;
         association->ShutdownCompleteCondition.broadcast();


         // ====== Generate "Communication Lost" control chunk =============
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_CommunicationLost;
            AssociationStatus dummy;
            SCTPSocketMaster::clearAssociationStatus(dummy);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),dummy);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),dummy);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::communicationErrorNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }
}


// ###### SCTP communication error notification callback ####################
void SCTPSocketMaster::communicationErrorNotif(unsigned int   assocID,
                                               unsigned short status,
                                               void*          dummy)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Communication Error Notification, Status=%d",assocID,status);
   cerr << str << endl;
#endif


   // ====== Generate "Communication Error" control chunk ==================
   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_CommunicationError;
            AssociationStatus status;
            SCTPSocketMaster::getAssociationStatus(assocID,0,status);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::communicationLostNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }
}


// ###### SCTP restart notification callback ################################
void SCTPSocketMaster::restartNotif(unsigned int assocID, void* ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Restart Notification",assocID);
   cerr << str << endl;
#endif


   // ====== Generate "Communication Error" control chunk ==================
   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID);
      if(association != NULL) {
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_Restart;
            AssociationStatus status;
            SCTPSocketMaster::getAssociationStatus(assocID,0,status);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),status);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::restartNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }
}


// ###### SCTP shutdown complete notification callback ######################
void SCTPSocketMaster::shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr)
{
#ifdef PRINT_NOTIFICATIONS
   char str[256];
   snprintf((char*)&str,sizeof(str),
               "A%04d: Shutdown Complete Notification",assocID);
   cerr << str << endl;
#endif


   SCTPSocket* socket = SCTPSocketMaster::getSocketForAssociationID(assocID);
   if(socket != NULL) {
      SCTPAssociation* association = socket->getAssociationForAssociationID(assocID,false);
      if(association != NULL) {
         // ====== Notify waiting main thread ===============================
         association->ShutdownCompleteNotification = true;
         association->ShutdownCompleteCondition.broadcast();

         // ====== Generate "Shutdown Complete" control chunk ==============
         AssociationControlData* acd = new AssociationControlData;
         if(acd != NULL) {
            acd->Type = AssociationControlData::ACDT_ShutdownComplete;
            AssociationStatus dummy;
            SCTPSocketMaster::clearAssociationStatus(dummy);
            if(socket->HasGlobalQueue) {
               socket->GlobalQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),dummy);
            }
            else {
               association->InQueue.addChunk(assocID,0,0,0,0,acd,sizeof(AssociationControlData),dummy);
            }
         }
         else {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::shutdownCompleteNotif() - Out of memory!"
                 << endl;
            exit(1);
         }
      }
   }
}


// ###### Lock callback #####################################################
void SCTPSocketMaster::lock(void* data)
{
   SCTPSocketMaster* task = (SCTPSocketMaster*)data;
   task->synchronized();
}


// ###### Unlock callback ###################################################
void SCTPSocketMaster::unlock(void* data)
{
   SCTPSocketMaster* task = (SCTPSocketMaster*)data;
   task->unsynchronized();
}


// ###### Get socket for given association ID ###############################
SCTPSocket* SCTPSocketMaster::getSocketForAssociationID(const unsigned int assocID)
{
   unsigned short instanceID = 0;
   if(sctp_getInstanceID(assocID,&instanceID) == 0) {
      if(instanceID != 0) {
         multimap<unsigned short, SCTPSocket*>::iterator iterator =
            SCTPSocketMaster::SCTPSocketList.find(instanceID);
         if(iterator != SCTPSocketMaster::SCTPSocketList.end()) {
            return(iterator->second);
         }
      }
#ifdef DEBUG
      else {
        cerr << "WARNING: SCTPSocketMaster::getSocketForAssociationID() - "
                "Instance ID for association ID #" << assocID << " is zero!" << endl;
      }
#endif
   }
#ifdef DEBUG
   else {
      cerr << "WARNING: SCTPSocketMaster::getSocketForAssociationID() - "
              "No instance ID for association ID #" << assocID << " found!" << endl;
   }
#endif
   return(NULL);
}


// ###### Clear AssociationStatus structure #################################
void SCTPSocketMaster::clearAssociationStatus(AssociationStatus& status)
{
   status.AssocID         = 0;
   status.StreamID        = 0;
   status.RemotePort      = 0;
   status.RemoteAddresses = 0;
   for(unsigned int i = 0;i < SCTP_MAX_NUM_ADDRESSES;i++) {
      status.RemoteAddress[i][0] = 0x00;
   }
}


// ###### Initialize AssociationStatus structure ############################
bool SCTPSocketMaster::getAssociationStatus(const unsigned int assocID,
                                            const unsigned int streamID,
                                            AssociationStatus& status)
{
   if(sctp_getAssocStatus(assocID,&status.Status) == 0) {
      status.AssocID  = assocID;
      status.StreamID = streamID;

      status.RemotePort      = status.Status.destPort;
      status.RemoteAddresses = min((unsigned short)SCTP_MAX_NUM_ADDRESSES,status.Status.numberOfAddresses);
      for(unsigned int i = 0;i < status.RemoteAddresses;i++) {
         if(sctp_getPathStatus(assocID,i,&status.PathStatus[i]) != 0) {
            cerr << "INTERNAL ERROR: SCTPSocketMaster::getAssociationStatus() - sctp_getPathStatus() failure!"
                 << endl;
            exit(1);
         }
         strcpy((char*)&status.RemoteAddress[i],(char*)&status.PathStatus[i].destinationAddress);
      }
      return(true);
   }
   SCTPSocketMaster::clearAssociationStatus(status);
   return(false);
}


// }
