

#ifndef SCTPSOCKETMASTER_H
#define SCTPSOCKETMASTER_H


#include "tdsystem.h"
#include "thread.h"
#include "sctpsocket.h"
#include "sctpchunkarrivalqueue.h"


#include <sctp.h>
#include <multimap.h>


// namespace Coral {


/**
  * This class manages the interaction between the SCTP Socket class and the
  * userland SCTP implementation. It is implemented as a singleton and
  * automatically instantiated at program startup.
  *
  * @short   SCTP Socket Master
  * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
  * @version 1.0
  */
class SCTPSocketMaster : public Thread
{
   // ====== Friend classes =================================================
   friend class SCTPSocket;
   friend class SCTPConnectionlessSocket;
   friend class SCTPAssociation;


   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   SCTPSocketMaster();

   /**
     * Destructor.
     */
   ~SCTPSocketMaster();


   // ====== Lock/unlock ====================================================
   /**
     * Lock the SCTP implementation for exclusive access. This includes
     * Thread's critical section lock.
     */
   inline void lock();

   /**
     * Unlock the SCTP implementation. This includes
     * Thread's critical section lock.
     */
   inline void unlock();


   // ====== Public data ====================================================
   public:
   /**
     * Result of sctp_initLibrary() call. If not 0, the SCTP initialization
     * failed.
     */
   static int InitializationResult;


   // ====== Protected data =================================================
   protected:
   static SCTPSocketMaster                      MasterInstance;
   static SCTP_ulpCallbacks                     SCTPCallbacks;
   static multimap<unsigned short, SCTPSocket*> SCTPSocketList;


   static SCTPSocket* getSocketForAssociationID(const unsigned int assocID);


   // ====== Private data ===================================================
   private:
   void run();
   static void lock(void* data);
   static void unlock(void* data);
   static void clearAssociationStatus(AssociationStatus& status);
   static bool getAssociationStatus(const unsigned int assocID,
                                    const unsigned int streamID,
                                    AssociationStatus& status);
   static void dataArriveNotif(unsigned int assocID,
                               unsigned int streamID,
                               unsigned int len,
                               unsigned int protoID,
                               unsigned int unordered,
                               void*        ulpDataPtr);
   static void networkStatusChangeNotif(unsigned int   assocID,
                                        short          destAddrIndex,
                                        unsigned short newState,
                                        void*          ulpDataPtr);
   static void* communicationUpNotif(unsigned int   assocID,
                                     unsigned short status,
                                     int            noOfDestinations,
                                     unsigned short noOfInStreams,
                                     unsigned short noOfOutStreams,
                                     void*          dummy);
   static void communicationLostNotif(unsigned int   assocID,
                                      unsigned short status,
                                      void*          ulpDataPtr);
   static void communicationErrorNotif(unsigned int   assocID,
                                       unsigned short status,
                                       void*          dummy);
   static void sendFailureNotif(unsigned int   assocID,
                                unsigned char* unsent_data,
                                unsigned int   dataLength,
                                unsigned int*  context,
                                void*          dummy);
   static void restartNotif(unsigned int assocID, void* ulpDataPtr);
   static void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr);
};


// }


#include "sctpsocketmaster.icc"


#endif
