
#include <config.h>

#include "sctpsocketwrapper.h"

#include "tdsystem.h"
#include "tdsocket.h"
#include "internetaddress.h"
#include "sctpsocket.h"
#include "sctpconnectionlesssocket.h"
#include "sctpassociation.h"


#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>


// using namespace Coral;


// ###### TDSocket structure ################################################
struct TDSocket
{
   enum TDSocketTypes {
      TDST_Invalid = 0,
      TDST_System  = 1,
      TDST_SCTP_CO = 2,
      TDST_SCTP_CL = 3
   };

   unsigned int Type;
   int          SubType;

   union TDSocketUnion {
      int SystemSocketID;
      struct SCTP {
         SCTPSocket*               SCTPCOSocketPtr;
         SCTPConnectionlessSocket* SCTPCLSocketPtr;
         SCTPAssociation*          SCTPAssociationPtr;
         bool                      Flags;
      } SCTPSocketDesc;
   } Socket;
};


// ###### Class for TDSocket management #####################################
class TDSocketMaster
{
   protected:
   TDSocketMaster();
   ~TDSocketMaster();

   public:
   inline static TDSocket* getSocket(const int id);
   static int setSocket(const TDSocket& newSocket);

   private:
   static TDSocketMaster MasterInstance;
   static const unsigned int MaxSockets = 1024;
   static TDSocket Sockets[MaxSockets];
};

TDSocket TDSocketMaster::Sockets[TDSocketMaster::MaxSockets];
TDSocketMaster TDSocketMaster::MasterInstance;


// ###### Constructor #######################################################
TDSocketMaster::TDSocketMaster()
{
   for(unsigned int i = 0;i < MaxSockets;i++) {
      Sockets[i].Type = TDSocket::TDST_Invalid;
   }
}


// ###### Destructor ########################################################
TDSocketMaster::~TDSocketMaster()
{
   for(unsigned int i = 0;i < MaxSockets;i++) {
      td_close(i);
   }
}


// ###### Get TDSocket for given ID #########################################
inline TDSocket* TDSocketMaster::getSocket(const int id)
{
   if((id < (int)MaxSockets) && (id > 0)) {
      return(&Sockets[id - 1]);
   }
   return(NULL);
}


// ###### Set TDSocket of given ID ##########################################
int TDSocketMaster::setSocket(const TDSocket& newSocket)
{
   for(int i = 0;i < (int)MaxSockets;i++) {
      if(Sockets[i].Type == TDSocket::TDST_Invalid) {
         Sockets[i] = newSocket;
         return(i + 1);
      }
   }
   return(-1);
}


// ###### socket() wrapper ##################################################
int td_socket(int domain, int type, int protocol)
{
   TDSocket tdSocket;
   if(protocol == IPPROTO_SCTP) {
      tdSocket.SubType                                  = type;
      tdSocket.Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
      tdSocket.Socket.SCTPSocketDesc.Flags              = 0;
      if(type == SOCK_STREAM) {
         SCTPSocket* sctpSocket = new SCTPSocket;
         if(sctpSocket == NULL) {
            return(-ENOSR);
         }
         tdSocket.Type                                  = TDSocket::TDST_SCTP_CO;
         tdSocket.Socket.SCTPSocketDesc.SCTPCOSocketPtr = sctpSocket;
         tdSocket.Socket.SCTPSocketDesc.SCTPCLSocketPtr = NULL;
      }
      else {
         SCTPConnectionlessSocket* sctpSocket = new SCTPConnectionlessSocket(false);
         if(sctpSocket == NULL) {
            return(-ENOSR);
         }
         tdSocket.Type                                  = TDSocket::TDST_SCTP_CL;
         tdSocket.Socket.SCTPSocketDesc.SCTPCLSocketPtr = sctpSocket;
         tdSocket.Socket.SCTPSocketDesc.SCTPCOSocketPtr = NULL;
      }
      return(TDSocketMaster::setSocket(tdSocket));
   }
   else {
      tdSocket.Type                  = TDSocket::TDST_System;
      tdSocket.SubType               = type;
      tdSocket.Socket.SystemSocketID = socket(domain,type,protocol);
      if(tdSocket.Socket.SystemSocketID >= 0) {
         return(TDSocketMaster::setSocket(tdSocket));
      }
      return(-1);
   }
}


// ###### close() wrapper ###################################################
int td_close(int sockfd)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               delete tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr;
               tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
            }
            if(tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr != NULL) {
               delete tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr;
               tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr = NULL;
            }
          break;
         case TDSocket::TDST_SCTP_CL:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               delete tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr;
               tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
            }
            if(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr != NULL) {
               delete tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr;
               tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr = NULL;
            }
          break;
         case TDSocket::TDST_System:
            close(tdSocket->Socket.SystemSocketID);
            tdSocket->Socket.SystemSocketID = 0;
          break;
         default:
            return(-ENXIO);
          break;
      }
      tdSocket->Type = TDSocket::TDST_Invalid;
      return(0);
   }
   return(-EBADF);
}


// ###### bind() wrapper ####################################################
int td_bind(int sockfd, struct sockaddr* my_addr, socklen_t addrlen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  return(-EBUSY);
               }
               InternetAddress address(my_addr,addrlen);
               address.setPrintFormat(InternetAddress::PF_Address|InternetAddress::PF_IPOnly);
               InternetAddress* addressPtr = &address;
               return(tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr->bind(address.getPort(),1,1,1,(const InternetAddress**)&addressPtr));
            }
          break;
         case TDSocket::TDST_SCTP_CL:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  return(-EBUSY);
               }
               InternetAddress address(my_addr,addrlen);
               address.setPrintFormat(InternetAddress::PF_Address|InternetAddress::PF_IPOnly);
               InternetAddress* addressPtr = &address;
               return(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr->bind(address.getPort(),1,1,1,(const InternetAddress**)&addressPtr));
            }
          break;
         case TDSocket::TDST_System:
             return(bind(tdSocket->Socket.SystemSocketID,my_addr,addrlen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### listen() wrapper ####################################################
int td_listen(int sockfd, int backlog)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr != NULL) {
               tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr->listen(backlog);
               return(0);
            }
            return(-EOPNOTSUPP);
          break;
         case TDSocket::TDST_SCTP_CL:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr != NULL) {
               tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr->listen(backlog);
               return(0);
            }
            return(-EOPNOTSUPP);
          break;
         case TDSocket::TDST_System:
            return(listen(tdSocket->Socket.SystemSocketID,backlog));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### accept() wrapper ####################################################
int td_accept(int sockfd, struct sockaddr* addr, socklen_t* addrlen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr != NULL) {
                  InternetAddress  remoteAddress;
                  InternetAddress* remoteAddressArray = &remoteAddress;
                  unsigned int     remoteAddresses = 1;

                  SCTPAssociation* association = tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr->accept(&remoteAddressArray,remoteAddresses);
                  if(association != NULL) {
                     TDSocket newTDSocket = *tdSocket;
                     newTDSocket.Socket.SCTPSocketDesc.SCTPCLSocketPtr    = NULL;
                     newTDSocket.Socket.SCTPSocketDesc.SCTPCOSocketPtr    = NULL;
                     newTDSocket.Socket.SCTPSocketDesc.SCTPAssociationPtr = association;
                     return(TDSocketMaster::setSocket(newTDSocket));
                  }
                  else {
                     return(0);
                  }
               }
               return(-EOPNOTSUPP);
            }
          break;
         case TDSocket::TDST_SCTP_CL:
            return(-EOPNOTSUPP);
          break;
         case TDSocket::TDST_System:
            {
               TDSocket newTDSocket = *tdSocket;
               newTDSocket.Socket.SystemSocketID = accept(tdSocket->Socket.SystemSocketID,addr,addrlen);
               if(newTDSocket.Socket.SystemSocketID >= 0) {
                  return(TDSocketMaster::setSocket(newTDSocket));
               }
               return(newTDSocket.Socket.SystemSocketID);
            }
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### getsockname() wrapper #############################################
int td_getsockname(int sockfd, struct sockaddr* name, socklen_t* namelen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            {
               InternetAddress  localAddress;
               InternetAddress* localAddressArray = &localAddress;
               unsigned int     localAddresses = 1;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getLocalAddresses(&localAddressArray,localAddresses);
               }
               else if(tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr->getLocalAddresses(&localAddressArray,localAddresses);
               }
               else if(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr->getLocalAddresses(&localAddressArray,localAddresses);
               }
               else {
                  return(-EBADF);
               }

               if(localAddresses >= 1) {
                  if(localAddress.getSystemAddress(name,*namelen,(InternetAddress::UseIPv6 == true) ? AF_INET6 : AF_INET) > 0) {
                     return(0);
                  }
                  return(-ENAMETOOLONG);
               }
               return(-ENXIO);
            }
          break;
         case TDSocket::TDST_System:
            return(getsockname(tdSocket->Socket.SystemSocketID,name,namelen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### getpeername() wrapper #############################################
int td_getpeername(int sockfd, struct sockaddr* name, socklen_t* namelen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            {
               InternetAddress  remoteAddress;
               InternetAddress* remoteAddressArray = &remoteAddress;
               unsigned int     remoteAddresses = 1;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getRemoteAddresses(&remoteAddressArray,remoteAddresses);
               }
               else {
                  return(-EBADF);
               }

               if(remoteAddresses >= 1) {
                  if(remoteAddress.getSystemAddress(name,*namelen,(InternetAddress::UseIPv6 == true) ? AF_INET6 : AF_INET) > 0) {
                     return(0);
                  }
                  return(-ENAMETOOLONG);
               }
               return(-ENXIO);
            }
          break;
         case TDSocket::TDST_SCTP_CL:
            return(-EOPNOTSUPP);
          break;
         case TDSocket::TDST_System:
            return(getpeername(tdSocket->Socket.SystemSocketID,name,namelen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### fcntl() wrapper ###################################################
int td_fcntl(int sockfd, int cmd, struct flock* lock)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               switch(cmd) {
                  case F_GETFL:
                     return(tdSocket->Socket.SCTPSocketDesc.Flags);
                    break;
                  case F_SETFL:
                      tdSocket->Socket.SCTPSocketDesc.Flags = (unsigned int)lock;
                      return(0);
                    break;
                  default:
                     return(-EOPNOTSUPP);
                   break;
               }
            }
            else {
               return(-EBADF);
            }
          break;
         case TDSocket::TDST_System:
            return(fcntl(tdSocket->Socket.SystemSocketID,cmd,lock));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### ioctl() wrapper ###################################################
int td_ioctl(int sockfd, int request, const void* argp)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            return(-EOPNOTSUPP);
          break;
         case TDSocket::TDST_System:
            return(ioctl(tdSocket->Socket.SystemSocketID,request,argp));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### getsockopt() wrapper ##############################################
int td_getsockopt(int sockfd, int level, int optname, void* optval, socklen_t* optlen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            {
               switch(level) {
                  // ====== IPv6 ============================================
                  case SOL_IPV6:
                      switch(optname) {
                         case IPV6_FLOWINFO:
                            return(0);
                          break;
                         case IPV6_FLOWINFO_SEND:
                            return(0);
                          break;
                         default:
                            return(-EOPNOTSUPP);
                          break;
                      }
                   break;
                  // ====== IPv4 ============================================
                  case SOL_IP:
                      switch(optname) {
                         case IP_TOS:
                            return(0);
                          break;
                         case IP_RECVTOS:
                            return(0);
                          break;
                         default:
                            return(-EOPNOTSUPP);
                          break;
                      }
                   break;
                  // ====== Default =========================================
                  default:
                     return(-EOPNOTSUPP);
                   break;
               }
            }
          break;
         case TDSocket::TDST_System:
            return(getsockopt(tdSocket->Socket.SystemSocketID,level,optname,optval,optlen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### setsockopt() wrapper ##############################################
int td_setsockopt(int sockfd, int level, int optname, const void* optval, socklen_t optlen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            {
               switch(level) {
                  // ====== IPv6 ============================================
                  case SOL_IPV6:
                      switch(optname) {
                         case IPV6_FLOWINFO:
                            return(0);
                          break;
                         case IPV6_FLOWINFO_SEND:
                            return(0);
                          break;
                         default:
                            return(-EOPNOTSUPP);
                          break;
                      }
                   break;
                  // ====== IPv4 ============================================
                  case SOL_IP:
                      switch(optname) {
                         case IP_TOS:
                            return(0);
                          break;
                         case IP_RECVTOS:
                            return(0);
                          break;
                         default:
                            return(-EOPNOTSUPP);
                          break;
                      }
                   break;
                  // ====== Default =========================================
                  default:
                     return(-EOPNOTSUPP);
                   break;
               }
            }
          break;
         case TDSocket::TDST_System:
            return(setsockopt(tdSocket->Socket.SystemSocketID,level,optname,optval,optlen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### shutdown() wrapper ################################################
int td_shutdown(int sockfd, int how)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->shutdown();
               return(0);
            }
            return(-EOPNOTSUPP);
          break;
         case TDSocket::TDST_System:
            return(shutdown(tdSocket->Socket.SystemSocketID,how));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### connect() wrapper #################################################
int td_connect(int sockfd, const struct sockaddr* serv_addr, socklen_t addrlen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  return(-EBUSY);
               }
               if(tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr != NULL) {
                  const InternetAddress address((sockaddr*)serv_addr,addrlen);
                  tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr = tdSocket->Socket.SCTPSocketDesc.SCTPCOSocketPtr->associate(1,address);
                  if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                     return(0);
                  }
               }
               return(-EOPNOTSUPP);
            }
          break;
         case TDSocket::TDST_SCTP_CL:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  return(-EBUSY);
               }
               if(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr != NULL) {
                  const InternetAddress address((sockaddr*)serv_addr,addrlen);
                  tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr = tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr->associate(1,address);
                  if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                     return(0);
                  }
               }
               return(-EOPNOTSUPP);
            }
          break;
         case TDSocket::TDST_System:
             return(connect(tdSocket->Socket.SystemSocketID,serv_addr,addrlen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### recv() wrapper ####################################################
int td_recv(int sockfd, void* buf, size_t len, int flags)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            {
               unsigned short streamID;
               unsigned int   protoID;
               int            newFlags      = flags;
               size_t         dataLength    = len;
               size_t         controlLength = 0;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  const int result =
                     tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->receive(
                        (char*)buf, dataLength,
                        NULL, controlLength,
                        newFlags, streamID, protoID);
                  if(dataLength > 0) {
                     return(dataLength);
                  }
                  return(result);
               }
               else {
                  return(-ENOTCONN);
               }
            }
          break;
         case TDSocket::TDST_System:
             return(recv(tdSocket->Socket.SystemSocketID,buf,len,flags));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### recvfrom() wrapper ################################################
int td_recvfrom(int sockfd, void* buf, size_t len, int flags,
                struct sockaddr* from, socklen_t* fromlen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            td_getpeername(sockfd,from,fromlen);
            return(td_recv(sockfd,buf,len,flags));
          break;
         case TDSocket::TDST_SCTP_CL:
            {
               unsigned short    streamID;
               unsigned int      protoID;
               int               newFlags      = flags;
               size_t            dataLength    = len;
               size_t            controlLength = 0;
               InternetAddress   remoteAddress;
               InternetAddress*  remoteAddressArray = &remoteAddress;
               unsigned int      remoteAddresses = 1;
               AssociationStatus status;

               if(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr != NULL) {
                  const int result =
                     tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr->receiveFrom(
                        (char*)buf, dataLength,
                        NULL, controlLength,
                        newFlags,
                        streamID, protoID,
                        &remoteAddressArray, remoteAddresses,
                        status);
                  if(result == 0) {
                     if(remoteAddresses >= 1) {
                        if(remoteAddress.getSystemAddress(from,*fromlen,(InternetAddress::UseIPv6 == true) ? AF_INET6 : AF_INET) > 0) {
                           return(dataLength);
                        }
                        return(-ENAMETOOLONG);
                     }
                     *fromlen = 0;
                     return(dataLength);
                  }
                  return(result);
               }
            }
          break;
         case TDSocket::TDST_System:
            return(recvfrom(tdSocket->Socket.SystemSocketID,buf,len,flags,from,fromlen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### recvmsg() wrapper #################################################
int td_recvmsg(int sockfd, struct msghdr* msg, int flags)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            {
               int result = td_recvfrom(sockfd,
                                        (sockaddr*)msg->msg_iov->iov_base,
                                        msg->msg_iov->iov_len,
                                        flags,
                                        (sockaddr*)msg->msg_name,
                                        &msg->msg_namelen);

               msg->msg_controllen = 0;
               return(result);
            }
          break;
         case TDSocket::TDST_System:
            return(recvmsg(tdSocket->Socket.SystemSocketID,msg,flags));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### send() wrapper ####################################################
int td_send(int sockfd, const void* msg, size_t len, int flags)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
         case TDSocket::TDST_SCTP_CL:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               return(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->send(
                         (char*)msg, len, flags,
                         0x0000, 0x00000000));
            }
            return(-ENOTCONN);
          break;
         case TDSocket::TDST_System:
            return(send(tdSocket->Socket.SystemSocketID,msg,len,flags));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### sendto() wrapper ####################################################
int td_sendto(int sockfd, const void* msg, size_t len, int flags,
              const struct sockaddr* to, socklen_t tolen)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            return(td_send(sockfd,msg,len,flags));
          break;
         case TDSocket::TDST_SCTP_CL:
            {
               const InternetAddress destination((sockaddr*)to,tolen);

               // ====== Try default connection, if address is invalid ======
               if((destination.isNull()) || (!destination.isValid())) {
                  return(td_send(sockfd,msg,len,flags));
               }
               else {
                  if(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr != NULL) {
                     return(tdSocket->Socket.SCTPSocketDesc.SCTPCLSocketPtr->sendTo(
                               (char*)msg, len, flags,
                               0x0000, 0x00000000,
                               destination));
                  }
               }
            }
          break;
         case TDSocket::TDST_System:
            return(sendto(tdSocket->Socket.SystemSocketID,msg,len,flags,to,tolen));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### sendmsg() wrapper #################################################
int td_sendmsg(int sockfd, const struct msghdr* msg, int flags)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case TDSocket::TDST_SCTP_CO:
            {
               return(td_send(sockfd,
                              (sockaddr*)msg->msg_iov->iov_base,
                              msg->msg_iov->iov_len,
                              flags));
            }
          break;
         case TDSocket::TDST_SCTP_CL:
            {
               const int result = td_sendto(sockfd,
                                     (sockaddr*)msg->msg_iov->iov_base,
                                     msg->msg_iov->iov_len,
                                     flags,
                                     (sockaddr*)msg->msg_name,
                                     msg->msg_namelen);
               return(result);
            }
          break;
         case TDSocket::TDST_System:
            return(sendmsg(tdSocket->Socket.SystemSocketID,msg,flags));
          break;
      }
      return(-ENXIO);
   }
   return(-EBADF);
}


// ###### read() wrapper ####################################################
ssize_t td_read(int fd, void* buf, size_t count)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(fd);
   if(tdSocket != NULL) {
      if(tdSocket->Type == TDSocket::TDST_System) {
         return(read(tdSocket->Socket.SystemSocketID,buf,count));
      }
      else {
         return(td_recv(fd,buf,count,0));
      }
   }
   return(-EBADF);
}


// ###### write() wrapper ####################################################
ssize_t td_write(int fd, const void* buf, size_t count)
{
   TDSocket* tdSocket = TDSocketMaster::getSocket(fd);
   if(tdSocket != NULL) {
      if(tdSocket->Type == TDSocket::TDST_System) {
         return(write(tdSocket->Socket.SystemSocketID,buf,count));
      }
      else {
         return(td_send(fd,buf,count,0));
      }
   }
   return(-EBADF);
}
