


#include "timedthread.h"
#include "tools.h"


#include <sys/time.h>
#include <signal.h>


// namespace Coral {


// ###### Constructor #######################################################
TimedThread::TimedThread(const card64 usec, const cardinal flags)
   : Thread(flags)
{
   Interval                = usec;
   TimerCorrection         = 10;
   FastStart               = true;
   AlarmHandlerInitialized = false;
   NewInterval             = false;
}


// ###### Destructor ########################################################
TimedThread::~TimedThread()
{
}


// ###### Set new interval ##################################################
void TimedThread::setInterval(const card64 usec)
{
   synchronized();

   Interval    = usec;
   NewInterval = true;

   // Send SIGUSR1 to thread to change interval
   if(running()) {
      if(pthread_self() != PThread) {
         if(AlarmHandlerInitialized == true) {
            pthread_kill(PThread,SIGUSR1);
         }
      }
   }

   unsynchronized();
}


// ###### Reimplementation of stop() ########################################
void TimedThread::stop()
{
   synchronized();
   AlarmHandlerInitialized = false;
   unsynchronized();
   Thread::stop();
}


// ###### The Thread's run() implementation #################################
void TimedThread::run()
{
   // Block SIGALRM and SIGUSR1 signal handlers
   synchronized();
   int      signalReceived;
   sigset_t signalSet;
   sigemptyset(&signalSet);
   sigaddset(&signalSet,SIGALRM);
   sigaddset(&signalSet,SIGUSR1);
   pthread_sigmask(SIG_BLOCK,&signalSet,NULL);
   AlarmHandlerInitialized = true;
   LeaveCorrectionLoop     = false;

   // Initialize timer to interval; we will receive a SIGALRM every Interval
   // microseconds
   itimerval itimer;
   if(FastStart == true) {
      itimer.it_value.tv_sec     = 0;
      itimer.it_value.tv_usec    = 1;
   }
   else {
      itimer.it_value.tv_sec     = Interval / 1000000;
      itimer.it_value.tv_usec    = Interval % 1000000;
   }
   itimer.it_interval.tv_sec  = Interval / 1000000;
   itimer.it_interval.tv_usec = Interval % 1000000;
   if(setitimer(ITIMER_REAL,&itimer,NULL) == -1) {
      cerr << "ERROR: TimedThread::run() - setitimer() failed!" << endl;
      exit(0);
   }

   // Get first time stamp for correction of inaccurate system timer
   card64   firstTimeStamp = getMicroTime();
   card64   interval       = Interval;
   cardinal count          = 0;   

   unsynchronized();

   // Main loop: sigwait() waits for SIGALRM and SIGUSR1 signals
   for(;;) {
      // Wait for signal and call timerEvent(), if SIGALRM has been received.
      sigwait(&signalSet,&signalReceived);
      if(signalReceived == SIGALRM) {
         const card64 correctedCount =
                         (getMicroTime() - firstTimeStamp) / interval;

         // ====== Call user's timerEvent() =================================
         if(TimerCorrection == 0) {   // With TimerCorrection enabled:
            timerEvent();             // Call is within correction loop!
            count++;
         }
         // ====== Correct inaccurate system timer ==========================
         else {
            int64 correction = (int64)correctedCount - (int64)count;
            if(correction == 0)
               correction = 1;
            if((correction > 0) && ((cardinal)correction < TimerCorrection)) {
               for(cardinal i = 0;i < (cardinal)correction;i++) {
                  timerEvent();
                  if(LeaveCorrectionLoop == true) {
                     break;
                  }
                  count++;
               }
            }
         }
         count               = correctedCount;
         LeaveCorrectionLoop = false;
      }

      // ====== Check, if Interval has to be updated =====================
      synchronized();
      if((NewInterval) && (Interval != interval)) {
         if(FastStart == true) {
            itimer.it_value.tv_sec  = 0;
            itimer.it_value.tv_usec = 1;
         }
         else {
            itimer.it_value.tv_sec     = Interval / 1000000;
            itimer.it_value.tv_usec    = Interval % 1000000;
         }

         itimer.it_interval.tv_sec  = Interval / 1000000;
         itimer.it_interval.tv_usec = Interval % 1000000;
         if(setitimer(ITIMER_REAL,&itimer,NULL) == -1) {
            cerr << "ERROR: TimedThread::run() - setitimer() failed!" << endl;
            exit(0);
         }
         firstTimeStamp = getMicroTime();
         count          = 0;
         interval       = Interval;
      }
      NewInterval = false;
      unsynchronized();
   }
}


// }
